/*
  File Name     : ft602_dclk_fifo.v 
  Department    : IC Design, FTDI SGP
  Author        : Do Ngoc Duong
  History       : 25 August 2016 - Initial Version 

  Description   : Dual clock FIFO control, 
                  LEN > 5 + 4*(wclk_fre/rclk_fre)  
*/
////////////////////////////////////////////////////////////////////////////////

module ft602_dclk_fifo (
//
  wrclk,
  wrrst_n,
  fifowr,
  wrfifoful,
  wrfifolen,
  wrmarklen,
  wrmark,
// 
  rdclk,
  rdrst_n,
  fiford,
  rdfifonempt,
  rdfifolen,
  rdmarklen,
  rdmark,
//  
  write,
  wraddr,
  read,
  rdaddr
);

////////////////////////////////////////////////////////////////////////////////
// Parameter declarations
parameter   ADB = 4;
parameter   LEN = 16;
////////////////////////////////////////////////////////////////////////////////
// Port declarations
input             wrclk;
input             wrrst_n;
input             fifowr;
output            wrfifoful;
output [ADB:0]    wrfifolen;
input  [ADB:0]    wrmarklen;
output            wrmark;
//
input             rdclk;
input             rdrst_n;
input             fiford;
output            rdfifonempt;
output [ADB:0]    rdfifolen;
input  [ADB:0]    rdmarklen;
output reg        rdmark;
//memory interface 
output            write;
output [ADB-1:0]  wraddr;
output            read;
output [ADB-1:0]  rdaddr;
////////////////////////////////////////////////////////////////////////////////
// Local logic and instantiation
//=== Across domain signals 
reg [ADB:0] rdpnt_gray;
reg [ADB:0] wrpnt_gray;
//************************ Write clock domain **********************************
reg [ADB:0] ftdi_sync_rdpnt_gray1;
reg [ADB:0] ftdi_sync_rdpnt_gray2;
//synchronize read pointer 
always @ ( posedge wrclk or negedge wrrst_n)
begin
  if(~wrrst_n)
  begin
    ftdi_sync_rdpnt_gray1  <= {(ADB+1){1'b0}};
    ftdi_sync_rdpnt_gray2  <= {(ADB+1){1'b0}};
  end
  else
  begin
    ftdi_sync_rdpnt_gray1  <= rdpnt_gray;
    ftdi_sync_rdpnt_gray2  <= ftdi_sync_rdpnt_gray1;
  end
end
//read pointer after decode Gray code  
reg [ADB:0]    drdpnt_bin;  
integer i; 
always @ (ftdi_sync_rdpnt_gray2)
begin
  for (i=0;i<(ADB+1);i=i+1)
    drdpnt_bin[i] = ^(ftdi_sync_rdpnt_gray2[ADB:0]>>i);
end
// 
reg [ADB:0]    wrpnt_bin;
wire           wrfifoful;
assign         wrfifoful   = (wrpnt_bin[ADB]^drdpnt_bin[ADB]) &
                             (wrpnt_bin[ADB-1:0] == drdpnt_bin[ADB-1:0]); 
//  
wire [ADB:0]   sublen;
assign         sublen      = wrpnt_bin[ADB:0] - drdpnt_bin[ADB:0]; 
//
wire [ADB:0]   wrfifolen;
assign         wrfifolen   = sublen;
//
wire           write;
assign         write       = fifowr & (~wrfifoful);
//              
wire [ADB-1:0] wraddr;
assign         wraddr      = wrpnt_bin[ADB-1:0];
// 
always @ ( posedge wrclk or negedge wrrst_n)
begin
  if(~wrrst_n)
    wrpnt_bin   <= {(ADB+1){1'b0}};
  else if (write)
    wrpnt_bin <= wrpnt_bin + 1'b1;
end
// Change wrcnt into gray code (so that only 1 bit change per posedge wclk)
// Ex: G_wrcnt[1:0] = 00 | 01 | 11 | 10
always @ ( posedge wrclk or negedge wrrst_n )
begin
  if(~wrrst_n)
    wrpnt_gray   <= {(ADB+1){1'b0}};
  else
    wrpnt_gray  <= wrpnt_bin ^ {1'b0,wrpnt_bin[ADB:1]};
end
//
reg  [ADB:0]   ftdi_sync_wrmarklen1;
reg  [ADB:0]   ftdi_sync_wrmarklen2;
always @ ( posedge wrclk or negedge wrrst_n)
begin
  if(~wrrst_n)
  begin 
    ftdi_sync_wrmarklen2 <= 'b0;
    ftdi_sync_wrmarklen1 <= 'b0;
  end 
  else 
  begin 
    ftdi_sync_wrmarklen2 <= ftdi_sync_wrmarklen1;
    ftdi_sync_wrmarklen1 <= wrmarklen - 3'd5;
  end
end
// 
reg [ADB:0]   sublen_p1;
always @ ( posedge wrclk or negedge wrrst_n)
begin
  if(~wrrst_n)
    sublen_p1 <= {(ADB+1){1'b0}};
  else 
    sublen_p1 <= sublen;
end 
//
assign wrmark = (sublen_p1 >= ftdi_sync_wrmarklen2);
/*
always @ ( posedge wrclk or negedge wrrst_n)
begin
  if(~wrrst_n)
    wrmark <= 1'b0;
  else 
    wrmark <= (sublen >= ftdi_sync_wrmarklen2) ; 
end 
*/ 
//************************** Read Clock Domain *********************************
//synchronize read pointer
reg [ADB:0]       ftdi_sync_wrpnt_gray1;
reg [ADB:0]       ftdi_sync_wrpnt_gray2;
always @ ( posedge rdclk or negedge rdrst_n )
begin
  if(~rdrst_n)
  begin
    ftdi_sync_wrpnt_gray1  <= {(ADB+1){1'b0}};
    ftdi_sync_wrpnt_gray2  <= {(ADB+1){1'b0}};
  end
  else
  begin
    ftdi_sync_wrpnt_gray1  <= wrpnt_gray;
    ftdi_sync_wrpnt_gray2  <= ftdi_sync_wrpnt_gray1;
  end
end
//Write pointer after decode Gray code
reg [ADB:0] dwrpnt_bin;    
reg [ADB:0] rdpnt_bin;    
always @ (ftdi_sync_wrpnt_gray2)
begin
  for (i=0;i<(ADB+1);i=i+1)
    dwrpnt_bin[i] = ^(ftdi_sync_wrpnt_gray2[ADB:0]>>i);
end
//
wire                rdfifonempt;
assign              rdfifonempt    = ~ (dwrpnt_bin[ADB:0] == rdpnt_bin[ADB:0]);
//
wire                read;
assign              read        = fiford & rdfifonempt;
//
wire [ADB-1:0]      rdaddr;
assign              rdaddr      = rdpnt_bin[ADB-1:0];
//
wire [ADB:0]        rdfifolen;
assign              rdfifolen   = dwrpnt_bin[ADB:0] - rdpnt_bin[ADB:0];
//
always @ ( posedge rdclk or negedge rdrst_n )
begin
  if(~rdrst_n)
    rdpnt_bin   <= {(ADB+1){1'b0}};
  else if (read) 
    rdpnt_bin <= rdpnt_bin + 1'b1;
end
// 
always @ ( posedge rdclk or negedge rdrst_n )
begin
  if(~rdrst_n)
    rdpnt_gray   <= {(ADB+1){1'b0}};
  else
    rdpnt_gray  <= rdpnt_bin ^ {1'b0,rdpnt_bin[ADB:1]};
end
//
//
reg  [ADB:0]   ftdi_sync_rdmarklen1;
reg  [ADB:0]   ftdi_sync_rdmarklen2;
always @ ( posedge rdclk or negedge rdrst_n )
begin
  if(~rdrst_n)
  begin 
    ftdi_sync_rdmarklen2 <= 'b0;
    ftdi_sync_rdmarklen1 <= 'b0;
  end 
  else 
  begin 
    ftdi_sync_rdmarklen2 <= ftdi_sync_rdmarklen1;
    ftdi_sync_rdmarklen1 <= rdmarklen;
  end
end
// 
always @ ( posedge rdclk or negedge rdrst_n )
begin
  if(~rdrst_n)
     rdmark <= 1'b0;
  else if ((rdfifolen == {{(ADB-1){1'b0}},{1'b1}}) & fiford)
     rdmark <= 1'b0;
  else if (rdfifolen > ftdi_sync_rdmarklen2)
     rdmark <= 1'b1;
end 
//
endmodule 
